
// RICM 3 - VM
// Encapsulation class for Volume data


#ifndef _Volume_H
#define _Volume_H

#include <iostream>
#include <stdlib.h>
#include <string>
#include <vector>

#include "V_3D.h"
#include "Matrix.h"


/// Container class for Structured Volume Data 
class Volume
{

 public :

  //************************************************//
  //************************************************//
  //CONSTRUCTORS
  //************************************************//
  //************************************************//

  /// Classical Constructor
  Volume();
  /// Classical Destructor
  ~Volume();

  /// Empty the data
  int destroy();

  //************************************************//
  //************************************************//
  //Load/save data
  //************************************************//
  //************************************************//

  /// Load v4d file
  int load_v4d(const std::string& filename );
  /// Load v4d file
  int load_v4d(const char* filename);

  /// Save v4d file
  int save_v4d(const std::string& filename ) const;
  /// Save v4d file
  int save_v4d(const char* filename) const;



  //************************************************//
  //************************************************//
  //2D Slice
  //************************************************//
  //************************************************//

  /// get the slice along the place defined by (p-x).n=0 with N1 x N2 evently sampled points
  /** p is assumed to be the center of the slice */
  /** dx, dy are the sampling distance */
  std::vector <double> slice(const V_3D& n,const V_3D& p,int N1,int N2,double dx,double dy) const;
  /// get the slice along the place defined by (p-x).n=0 with N1 x N2 evently sampled points
  /** p is assumed to be the center of the slice */
  /** Fill also the geometry of the plane*/
  int slice(const V_3D& n,const V_3D& p,int N1,int N2,double dx,double dy,std::vector <double>* vertex,std::vector <int>* connectivity,std::vector <double>* color) const;



  //************************************************//
  //************************************************//
  //Mesh and picture helper
  //************************************************//
  //************************************************//

  /// to export 2D-density in black and white in ppm (not really linked to Volume class, should create a special class for it)
  static int export_ppm(const std::string& filename,const std::vector <double>& current_slice,int N1,int N2);
  /// to export 2D-density in black and white in ppm (not really linked to Volume class, should create a special class for it)
  static int export_ppm(const char* filename,const std::vector <double>& current_slice,int N1,int N2);


  /// to export a black & white mesh (not really linked to Volume class, should create a special class for it)
  static int export_coff(const std::string& filename,const std::vector <double>& vertex,const std::vector <int>& connectivity,const std::vector <double>& color);
  /// to export a black & white mesh (not really linked to Volume class, should create a special class for it)
  static int export_coff(const char* filename,const std::vector <double>& vertex,const std::vector <int>& connectivity,const std::vector <double>& color);
  /// to export a mesh (not really linked to Volume class, should create a special class for it)
  static int export_off(const std::string& filename,const std::vector <double>& vertex,const std::vector <int>& connectivity);
  /// to export a black & white mesh (not really linked to Volume class, should create a special class for it)
  static int export_off(const char* filename,const std::vector <double>& vertex,const std::vector <int>& connectivity);


  //************************************************//
  //************************************************//
  //Isosurface
  //************************************************//
  //************************************************//

  /// marching cube algorithm.
  int isosurface(double value,std::vector <double>* vertex,std::vector <int>* connectivity) const;


  //************************************************//
  //************************************************//
  //Access
  //************************************************//
  //************************************************//

  /// get data
  double get(int k_x,int k_y,int k_z) const;
  /// get data
  double& get(int k_x,int k_y,int k_z);
  /// get vector data
  const std::vector<double>& get() const;
  /// get vector data
  std::vector <double>& get();

  /// get interpolated data
  /** return 0 if (x,y,z) ouside the data definition */
  double get(double x,double y,double z) const;
  /// same as get(double double double)
  double get_interpolated(double x,double y,double z) const;

  /// get size
  int get_size(int k_dim);

  /// resample the data
  int resample(double d_L);


 private :

  /// build a triangulated polygon for one cell for the isosurface
  int create_polygon(const std::vector <double>& cell_values,double value,int type_of_cube,V_3D position,std::vector <double>* vertex,std::vector <int>* connectivity) const;

  /// add the vertex in the complete mesh
  int add_vertex_in_mesh(std::vector <double>* vertex,std::vector <int>* connectivity,const V_3D& x) const;

  /// Vector of data of size N1 x N2 x N3
  /** access an element with data[k1+N1*k2+N1*N2*k3] */
  std::vector <double> data;
  /// size
  int N[3];


  /// helper indexation for the marching cube
  static int EdgeTable[256];
  static int TriangulationTable[256][16];

};

#endif
